package net.jaekl.cfb.db.driver;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;

import net.jaekl.cfb.db.Column;
import net.jaekl.cfb.db.Condition;
import net.jaekl.cfb.db.Row;
import net.jaekl.cfb.db.Sequence;
import net.jaekl.cfb.db.SequenceMock;
import net.jaekl.cfb.db.Sort;
import net.jaekl.cfb.db.Table;
import net.jaekl.cfb.db.TableMock;
import net.jaekl.cfb.db.TypeMismatchException;
import net.jaekl.cfb.util.Util;

public class DbDriverMock extends DbDriver {

	private HashMap<String, TableMock> m_tables;
	private HashMap<String, SequenceMock> m_sequences;
	
	public DbDriverMock() {
		super();
		m_tables = new HashMap<String, TableMock>();
		m_sequences = new HashMap<String, SequenceMock>();
	}

	@Override
	public void load() throws ClassNotFoundException {
		// no-op
	}

	@Override
	public Connection connect(String host, int port, String dbName, String user, String pass) 
		throws SQLException 
	{
		return null;
	}
	
	@Override
	public boolean createTable(Connection con, Table table) throws SQLException
	{
		assertNotNull(con);
		assertNotNull(table);
		assertFalse(m_tables.containsKey(table.getName()));
		
		TableMock tm = new TableMock(table);
		m_tables.put(table.getName(), tm);
		return true;
	}
	
	@Override
	public void dropTable(Connection con, Table table) throws SQLException 
	{
		assertNotNull(con);
		assertNotNull(table);
		assertTrue(table instanceof TableMock);
		
		if (m_tables.containsKey(table.getName())) {
			m_tables.remove(table.getName());
		}
		else {
			throw new SQLException("Table " + table.getName() + " does not exist.");
		}
	}
	
	@Override
	public boolean createSequence(Connection con, Sequence seq) throws SQLException 
	{
		assertNotNull(con);
		assertNotNull(seq);
		assertFalse(m_sequences.containsKey(seq.getName()));
		
		SequenceMock sm = (seq instanceof SequenceMock) ? (SequenceMock)seq : new SequenceMock(seq);
		
		m_sequences.put(seq.getName(), sm);
		return true;
	}
	
	public void dropSequence(Connection con, Sequence seq) throws SQLException
	{
		assertNotNull(con);
		assertNotNull(seq);
		assertTrue(seq instanceof SequenceMock);

		if (m_sequences.containsKey(seq.getName())) {
			m_sequences.remove(seq.getName());
		}
		else {
			throw new SQLException("Sequence " + seq.getName() + " does not exist.");
		}
	}
	
	public List<Row> select(Connection con, Column[] columns, Table[] tables, Condition[] conditions, Sort[] sorts, int limit)
		throws SQLException
	{
		assertNotNull(con);
		assertNotNull(columns);
		assertNotNull(tables);
		
		if (1 != tables.length) {
			throw new UnsupportedOperationException("DbDriverMock does not (yet) implement table joins");
		}
		
		Table schemaTable = tables[0];
		TableMock table = m_tables.get(schemaTable.getName());
		if (null == table) {
			throw new SQLException("Could not find table \"" + schemaTable.getName() + "\" in schema.");
		}
		
		ArrayList<Row> rows = table.mock_getRows();
		if (null != conditions) {
			ArrayList<Row> filteredRows = new ArrayList<Row>();
			for (Row row : rows) {
				boolean match = true;
				for (Condition condition : conditions) {
					try {
						if (!conditionSatisfied(condition, row)) {
							match = false;
						}
					}
					catch (TypeMismatchException exc) {
						throw new SQLException(exc);
					}
				}
				if (match) {
					filteredRows.add(row);
				}
			}
			rows = filteredRows;
		}
		
		ArrayList<Row> result = new ArrayList<Row>();
		
		// Now, convert the rows from select(*) to select(specified columns)
		for (Row row : rows) {
			Object[] values = new Object[columns.length];
			
			for (int outIdx = 0; outIdx < columns.length; ++outIdx) {
				for (int inIdx = 0; inIdx < row.getNumColumns(); ++inIdx) {
					if (row.getColumn(inIdx).getName().equals(columns[outIdx].getName())) {
						values[outIdx] = row.getValue(inIdx);
					}
				}
			}
			result.add(new Row(columns, values));
		}
		
		return result;
	}
	
	private boolean aLessThanB(Object a, Object b) 
	{
		if ((null == a) || (null == b)) {
			return false;
		}
		
		if (a instanceof Number) {
			if (! (b instanceof Number)) {
				throw new UnsupportedOperationException("Incompatible types");
			}
			Number first = (Number)a;
			Number second = (Number)b;
			return first.doubleValue() < second.doubleValue();
		}
		
		if (a instanceof Date) {
			if (! (b instanceof Date)) {
				throw new UnsupportedOperationException("Incompatible types");
			}
			Date first = (Date)a;
			Date second = (Date)b;
			int comp = first.compareTo(second);
			boolean result = (comp < 0);
			
			return (result);
		}
		
		throw new UnsupportedOperationException("Incompatible types");
	}
	
	private boolean conditionSatisfied(Condition condition, Row row) throws TypeMismatchException, SQLException
	{
		for (int idx = 0; idx < row.getNumColumns(); ++idx) {
			Column col = row.getColumn(idx);
			if (condition.getColumn().equals(col)) {
				switch(condition.getOperation()) {
				case EQUAL:
					return Util.objsAreEqual(condition.getValue(), row.getValue(idx));
				case GREATER_THAN:
					return (aLessThanB(condition.getValue(), row.getValue(idx)));
				case LESS_THAN:
					return (aLessThanB(row.getValue(idx), condition.getValue()));
				case NOT_NULL:
					return (null != row.getValue(idx));
				case NULL:
					return (null == row.getValue(idx));
				default:
					throw new UnsupportedOperationException("This condition operation is not supported by DbDriverMock");
				}
			}
		}
		throw new SQLException("Could not locate column \"" + condition.getColumn().getName() + "\" in TableMock.");
	}
	
	// Returns the number of rows inserted
	@Override
	public int insert(Connection con, Table table, Object[][] values) throws SQLException 
	{
		assertNotNull(con);
		assertNotNull(table);
		assertNotNull(values);
		
		TableMock tm = m_tables.get(table.getName());
		if (null == tm) {
			throw new SQLException("Table " + table.getName() + " does not exist.");
		}
		Column[] columns = tm.mock_getColumns();

		for (Object[] rowVals : values) {
			assertNotNull(rowVals);
			assertEquals(table.getNumColumns(), rowVals.length);
			
			tm.mock_insert(new Row(columns, rowVals));
		}
		
		return values.length;
	}


	public long nextVal(Connection con, Sequence seq) throws SQLException
	{
		assertNotNull(con);
		assertNotNull(seq);
		
		SequenceMock sm = m_sequences.get(seq.getName());
		return sm.mock_nextVal();
	}
}
